/*
 * Copyright 1993-2009 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and
 * international Copyright laws.  Users and possessors of this source code
 * are hereby granted a nonexclusive, royalty-free license to use this code
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOURCE CODE.
 *
 * U.S. Government End Users.   This source code is a "commercial item" as
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of
 * "commercial computer  software"  and "commercial computer software
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995)
 * and is provided to the U.S. Government only as a commercial end item.
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
 * source code with only those rights set forth herein.
 *
 * Any use of this source code in individual and commercial software must
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */

/*--------------------------------- Includes --------------------------------*/

#ifndef CUDADEBUGGER_H
#define CUDADEBUGGER_H

#include <cuda.h>

#if defined(__STDC__)
#include <inttypes.h>
#include <stdbool.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif


/* Windows isn't C99 compliant. Implement the subset we need */

#if defined(_WIN32) && !defined(_WIN64)
/* Windows 32-bit */
typedef          __int64   int64_t;
typedef unsigned __int64   uint64_t;
#define PRIxPTR "I32x"
#endif

#if defined(_WIN64)
/* Windows 64-bit */
typedef          long long int64_t;
typedef unsigned long long uint64_t;
#define PRIxPTR "I64x"
#endif

#if defined(_WIN32)
/* Windows 32- and 64-bit */
#define PRIx64  "I64x"
typedef short              int16_t;
typedef unsigned short     uint16_t;
typedef int                int32_t;
typedef unsigned           uint32_t;
typedef char               int8_t;
typedef unsigned char      uint8_t;
typedef unsigned char      bool;
#undef false
#undef true
#define false           0
#define true            1
#endif


#ifndef stdRangeMap_INCLUDED
typedef void *stdRangeMap_t;
typedef uint64_t rangemapRange_t;
typedef struct stdRange {
    rangemapRange_t start;
    rangemapRange_t length;
} rangemapDomain_t;
#endif

/*-------------------------- Exported Memory Segments ------------------------*/

/* The following are exported address class segments used by a debugging
   client (ex: cuda-gdb).  These values correspond directly with physical
   memory segments on the GPU. */

#define ptxStorageKindIterate(macro)          \
    macro(ptxCodeStorage,       "code"      ) \
    macro(ptxRegStorage,        "register"  ) \
    macro(ptxSregStorage,       "sregister" ) \
    macro(ptxConstStorage,      "constant"  ) \
    macro(ptxGlobalStorage,     "global"    ) \
    macro(ptxLocalStorage,      "local"     ) \
    macro(ptxParamStorage,      "parameter" ) \
    macro(ptxSharedStorage,     "shared"    ) \
    macro(ptxSurfStorage,       "surface"   ) \
    macro(ptxTexStorage,        "texture"   ) \
    macro(ptxTexSamplerStorage, "texsampler")

#define __ptxKindNameMacro(x,y) x,

typedef enum {
    ptxUNSPECIFIEDStorage,
    ptxStorageKindIterate( __ptxKindNameMacro )
    ptxMAXStorage
} ptxStorageKind;

#define PTX_NVSEGMENT_NAME(x)       ".nv"#x"Segment"
#define PTX_DEBUGINFO_NAME(key)     "__gpud__DebugInfo_"#key
#define PTX_ENDLABEL_NAME(entry)    "__$endLabel$__"#entry
#define PTX_SOURCELINK_SEGMENT_NAME ".nvSourceLinkSegment"
#define PTX_FATBIN_SEGMENT_NAME     ".nvFatBinSegment"

/*---------------------- FatFormat Debug Info Structure ----------------------*/

typedef struct CUDBGcode_table_st {
    uintptr_t                    virt;
    uintptr_t                    phys;
} *CUDBGcode_table;

typedef struct CUDBGsection_Table_st {
    uintptr_t                    start;
    uintptr_t                    end;
} *CUDBGsection_table;

typedef struct CUDBGphysical_location_st {
    char *                       name;
    unsigned                     bank;
    uint64_t                     offset;
} *CUDBGphysical_location;

typedef struct CUDBGvirtual_location_st {
    char *                       name;
    uint32_t                     size;
    uintptr_t                    address;
} *CUDBGvirtual_location;

typedef struct CUDBGdebug_info_st {
    uintptr_t                   *locationAddresses;
    CUDBGsection_table           sectionTable;
    CUDBGvirtual_location        variableRangeTable;
} *CUDBGdebug_info;

typedef struct CUDBGdevice_memory_descriptor_st {
    uint32_t                     bank;
    uint32_t                     size;
    uint64_t                     devicePointer;
    uintptr_t                    hostAddress;
} *CUDBGdevice_memory_descriptor;

typedef struct CUDBGmemory_segment_st {
    uint64_t                     vaddr;
    uint64_t                     offset;
    uint64_t                     size;

    /* Original handle */
    uint32_t mem;

    /* The dup'd handle */
    uint32_t newmem;

    /* Timestamp of last use */
    uint64_t lastUsed;
} CUDBGmemory_segment;

/*
 * The deviceMem array is not dynamically sized due to the way it's
 * passed from the work process to the debugger. This could be
 * improved. In the meantime, deviceMem[] is sized much larger than is
 * expected necessary.
 */
#define CUDBG_VIDMEMHEAP_MAX 256

/* Offsets into the memory map for various memory spaces */
typedef struct {
    int deviceMemLength;
    CUDBGmemory_segment deviceMem[CUDBG_VIDMEMHEAP_MAX];
    CUDBGmemory_segment funcMem;
    CUDBGmemory_segment localMem;
    CUDBGmemory_segment stackMem;

    // Relocation table
    uint32_t r_dptr, r_bytes;

    // Handles passed on from the driver
    uint32_t client;
    uint32_t device;
} CUDBGmemory_remapping;

/*--------------------------- Debugger System Calls --------------------------*/

/*
 * Names of interface functions that the
 * debugger runtime support in
 * the application must define:
 */
#define CUDBG_IPC_FLAG_NAME   cudbgIpcFlag
#define CUDBG_IPC_CALL_NAME   cudbgIpcCall
#define CUDBG_IPC_ARGS_NAME   cudbgIpcArgs


/*
 * Message layout used in system calls to debugger.
 * setting up and sending such messages is wrapped by
 * the Debug Client API functions specified below:
 */
   typedef enum {
       dbgRegisterClient,
       dbgRegisterModule,
       dbgRegisterFunction,
       dbgRegisterModuleConsts,
       dbgRegisterFunctionConsts,
       dbgReportMalloc,
       dbgReportFree,
       dbgReportDeviceFunctionLaunch,
       dbgReportDeviceExecutionLaunch,
       dbgReportDeviceExecutionStop,
       dbgSignalModuleComplete,
       dbgReportDevices,
   } CUDBGsyscall_kind;

typedef struct {
    uint32_t kind;

    union {
        struct {
            uintptr_t                    handle;
            uintptr_t                    thOffset;
            uint32_t                     threadId;
            uint32_t                     device;
            char                         deviceType[32];
            char                         smType[32];
        } registerClient;

        struct {
            uintptr_t                    handle;
            uint32_t                     threadId;
            CUDBGsection_table           sectionTable;
            uint32_t                     fatBinFlags;
            uint32_t                     funcMemHandle;
            uint64_t                     funcMemSize;
            char                        *stringTable;
            uint32_t                     stringTableSize;
            CUDBGdevice_memory_descriptor constTable;
            uint32_t                     constTableSize;
            CUDBGphysical_location       globalTable;
            uint32_t                     globalTableSize;
            CUDBGvirtual_location        variableRangeTable;
            uint32_t                     variableRangeTableSize;
            uint32_t                     nameIndex;
        } registerModule;

        struct {
            uintptr_t                    handle;
            uintptr_t                    module;
            uintptr_t                    entry;
            uint32_t                     amount;
            CUDBGcode_table              codeTable;
            CUDBGdevice_memory_descriptor constTable;
            uint32_t                     constTableSize;
            CUDBGphysical_location       variableTable;
            uint32_t                     variableTableSize;
            uint32_t                     nrofRegisters;
            uint32_t                     stackSize;
            uint32_t                     localSize;
            uint32_t                     sharedSize;
            uintptr_t                    stackAddress;
            uintptr_t                    localAddress;
            uintptr_t                    sharedAddress;
            uint32_t                     nameIndex;
        } registerFunction;

        struct {
            uintptr_t                    client;
            uintptr_t                    functionWrapper;
        } deviceFunctionLaunch;

        struct {
            uintptr_t                    client;
            uintptr_t                    function;
            CUDBGmemory_remapping        memMap;
            unsigned                     gridDim[2];
            unsigned                     blockDim[3];
            uint16_t                     ctaFlagsCookie;
        } deviceExecutionLaunch;

        struct {
            uintptr_t                    client;
            uintptr_t                    function;
        } deviceExecutionStop;

        struct {
            uintptr_t                    handle;
            uint64_t                     devicePointer;
            uint64_t                     size;
            uintptr_t                    hostAddress;
        } reportMalloc;

        struct {
            uintptr_t                    handle;
            uint64_t                     devicePointer;
        } reportFree;

        struct {
            uintptr_t                    module;
        } signalModuleComplete;

        struct {
            uint32_t total;
            uint32_t excluded;
        } reportDevices;

    } cases;
} CUDBGcommand;


/*------------------------ Breakpoint Thread Masks API -----------------------*/

typedef struct CUDBGcoordinates {
    unsigned bX, bY;
    unsigned tX, tY, tZ;
} *CUDBGcoordinates;

typedef struct CUDBGthread_and_block_coord {
    unsigned blockNr;
    unsigned threadNr;
} *CUDBGthread_and_block_coord;

typedef bool (*CUDBGthread_function)(CUDBGthread_and_block_coord n, CUDBGcoordinates c, void *data);

/*
 * Only PC for now:
 */
typedef struct {
    struct CUDBGcoordinates    thread;
    uintptr_t                      functionWrapper;
    uintptr_t                      pc;
    uintptr_t                      physPc;
} CUDBGcontext;

/*--------------------------------- Exports --------------------------------*/

/*
 * Move out to cuda interface module when things are stable:
 */

/*
 * We only export a single function, which populates a structs with function pointers
 */
typedef const struct CUDBGAPI_st *CUDBGAPI;

#define CUDBG_CURRENT_API 20090303

/*
 * nvGpuDebugExportAPI exports the debugging API in the passed in api
 * variable. For convenience, the cudbgAPI variable will get a copy of
 * the result. (api can be NULL in which case it's ignored)>
 *
 * Returns true upon success
 */
CUDBGAPI cudbgAPI;
bool cudbgExportAPI(unsigned requestedAPI, CUDBGAPI *api);

struct CUDBGAPI_st {
    bool             (*threadsTraverse)(unsigned tid, CUDBGthread_function f, void *data);
    bool             (*threadEqual)(CUDBGcoordinates t1, CUDBGcoordinates t2);
    bool             (*threadMatch)(CUDBGcoordinates mask, CUDBGcoordinates thread);
    bool             (*currentThreadMatch)(unsigned tid, CUDBGcoordinates mask);
    char *           (*getCoordsRepresentation)(bool isMask, CUDBGcoordinates mask);
    bool             (*constructCoord)(bool isMask, char *blockRepr, char *threadRepr, CUDBGcoordinates *mask);
    bool             (*setCurrentCoords)(unsigned tid, CUDBGcoordinates mask);
    CUDBGcoordinates (*getCurrentCoords)(unsigned tid);
    void             (*deleteCoords)(CUDBGcoordinates mask);
    void             (*initialize)(void);
    void             (*cleanup)(void);
    void             (*suspend)(void);
    void             (*resume)(unsigned pid, unsigned sstep, bool resumeall);
    int              (*receiveMessage)(CUDBGcommand *message);
    uintptr_t        (*denormalizeAddress)(unsigned pid, uintptr_t address, unsigned addressClass);
    uintptr_t        (*normalizeAddress)(unsigned pid, uintptr_t address, unsigned addressClass);
    bool             (*isPtxAddress)(uintptr_t address);
    uintptr_t        (*lowerPc)(uintptr_t address);
    bool             (*setBreakpoint)(uintptr_t address, CUDBGcoordinates thread, bool set);
    unsigned         (*readMemory)(uintptr_t address, void *buffer, unsigned amount);
    unsigned         (*writeMemory)(uintptr_t address, void *buffer, unsigned amount);
    bool             (*getContext)(unsigned tid, CUDBGcontext *context);
    char *           (*getState)(char *spec);
    void             (*deleteState)(char *state);
    char *           (*getMemoryState)(char *spec);
    void             (*deleteMemoryState)(char *state);
    bool             (*getInfo)(uintptr_t address, char **deviceType, char **smType);
    stdRangeMap_t    (*rangemapCreate)(void);
    void             (*rangemapDelete)(stdRangeMap_t rangemap);
    void             (*rangemapDefine)(stdRangeMap_t rangemap, rangemapDomain_t dom, void *ran);
    void *           (*rangemapApply)(stdRangeMap_t rangemap, rangemapRange_t dom);
    void             (*createExternSymbol)(char *symName);
    uintptr_t        (*lookupExternSymbol)(char *symName);
};

#ifdef __cplusplus
}
#endif

#endif
